# -------------------------------------------------------------------------------
# Copyright (c) 2018, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

"""
Created on March 1, 2018

@author: Craig Allwardt
"""
import base64
import inspect
import json
import logging
import os
import random
import threading
from collections import defaultdict
from datetime import datetime
from enum import Enum
from logging import Logger
from queue import Queue

from stomp import Connection12 as Connection
from stomp.exception import NotConnectedException
from time import sleep

_log: Logger = logging.getLogger(inspect.getmodulename(__file__))


class GRIDAPPSD_ENV_ENUM(Enum):
    GRIDAPPSD_USER = "GRIDAPPSD_USER"
    GRIDAPPSD_PASSWORD = "GRIDAPPSD_PASSWORD"
    GRIDAPPSD_ADDRESS = "GRIDAPPSD_ADDRESS"
    GRIDAPPSD_PORT = "GRIDAPPSD_PORT"
    GRIDAPPSD_PASS = "GRIDAPPSD_PASSWORD"


class TimeoutError(Exception):
    pass


class GOSS(object):
    """ Base class providing connections to a GOSS instance via stomp protocol
    """

    def __init__(self, username=None, password=None,
                 stomp_address='localhost', stomp_port='61613',
                 attempt_connection=True,
                 override_threading=None, stomp_log_level=logging.WARNING,
                 goss_log_level=logging.INFO):

        logging.getLogger('stomp.py').setLevel(stomp_log_level)
        logging.getLogger('goss').setLevel(goss_log_level)

        self.__user__ = username
        self.__pass__ = password
        self.stomp_address = stomp_address
        self.stomp_port = stomp_port

        # Environmental variables should overrule the passed arguments.
        if os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_USER.value):
            _log.debug(f"Environment for {GRIDAPPSD_ENV_ENUM.GRIDAPPSD_USER.value} is set")
            self.__user__ = os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_USER.value)

        if os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PASSWORD.value):
            _log.debug(f"Environment for {GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PASSWORD.value} is set")
            self.__pass__ = os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PASSWORD.value)

        if os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_ADDRESS.value):
            _log.debug(f"Environment for {GRIDAPPSD_ENV_ENUM.GRIDAPPSD_ADDRESS.value} is set")
            self.stomp_address = os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_ADDRESS.value)

        if os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PORT.value):
            _log.debug(f"Environment for {GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PORT.value} is set")
            self.stomp_port = os.environ.get(GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PORT.value)

        if not self.__user__ or not self.__pass__:
            raise ValueError("Invalid username/password specified.")
        self._conn = None
        self._ids = set()
        self._topic_set = set()
        self._override_thread_fc = override_threading
        self._router_callback = CallbackRouter()
        self.result_format = None
        self.__token = None

        if attempt_connection:
            self._make_connection()

    @property
    def connected(self):
        return self._conn is not None and self._conn.is_connected()

    def connect(self):
        self._make_connection()

    def disconnect(self):
        if self._conn:
            self._conn.disconnect()

        self._conn = None

    def override_threading(self, callback):
        self._override_thread_fc = callback

    def send(self, topic, message):
        self._make_connection()
        if isinstance(message, list) or isinstance(message, dict):
            message = json.dumps(message)
        _log.debug("Sending topic: {} body: {}".format(topic, message))
        self._conn.send(body=message, destination=topic,
                        headers={'GOSS_HAS_SUBJECT': True,
                                  'GOSS_SUBJECT': self.__token})
                                 
    def get_response(self, topic, message, timeout=5):
        id = datetime.now().strftime("%Y%m%d%h%M%S")
        reply_to = "/temp-queue/response.{}".format(id)
        
        if isinstance(message, str):
            message = json.loads(message)
        
        if 'resultFormat' in message:
            self.result_format = message['resultFormat'] 

        # Change message to string if we have a dictionary.
        if isinstance(message, dict):
            message = json.dumps(message)
        elif isinstance(message, list):
            message = json.dumps(message)
            
        class ResponseListener(object):
            def __init__(self, topic, result_format):
                self.response = None
                self._topic = topic
                self.result_format = result_format

            def on_message(self, header, message):

                _log.debug("Internal on message is: {} {}".format(header, message))
                try:
                    if self.result_format == 'JSON':
                        if isinstance(message, dict):
                            self.response = message
                        else:
                            self.response = json.loads(message)
                    else:   
                        self.response = message
                except ValueError:
                    self.response = dict(error="Invalid json returned",
                                         header=header,
                                         message=message)

            def on_error(self, headers, message):
                _log.error("ERR: {}".format(headers))
                _log.error("OUR ERROR: {}".format(message))

            def on_disconnect(self, header, message):
                _log.debug("Disconnected")

        listener = ResponseListener(reply_to, self.result_format)
        self.subscribe(reply_to, listener)

        self._conn.send(body=message, destination=topic,
                        headers={'reply-to': reply_to, 'GOSS_HAS_SUBJECT': True,
                                 'GOSS_SUBJECT': self.__token})
        count = 0

        while count < timeout:
            if listener.response is not None:
                break

            sleep(1)
            count += 1

        if listener.response is not None:
            return listener.response

        raise TimeoutError("Request not responded to in a timely manner!")

    def subscribe(self, topic, callback):
        """Subscribe to a given topic, and call callback on message.

        :param topic: topic to subscribe to. See topics.py.
        :param callback: function (callable) or class to be hit on
            every message. Note the class must have an "on_message"
            method. The function (or class's on_message method) will be
            passed two arguments: header and message.
        """
        conn_id = str(random.randint(1, 1000000))
        while conn_id in self._ids:
            conn_id = str(random.randint(1, 1000000))

        if not callback:
            err = "Invalid callback specified in subscription"
            _log.error(err)
            raise AttributeError(err)
        if not topic:
            err = "Invalid topic specified in subscription"
            _log.error(err)
            raise AttributeError(err)

        self._make_connection()

        if self._conn.get_listener('gridappsd') is None:
            self._conn.set_listener('gridappsd', self._router_callback)

        if callable(callback):
            self._router_callback.add_callback(topic, callback)
            # Handle the case where callback is a function
            # self._conn.set_listener(conn_id,
            #                         CallbackWrapperListener(callback, conn_id))
        else:
            # Case where the callback is (supposedly) a class.
            if not hasattr(callback, 'on_message'):
                m = "The given callback must have an 'on_message' method!"
                raise AttributeError(m)

            if not callable(callback.on_message):
                m = "The given callback's 'on_message' attribute must be " \
                    "callable!"
                raise TypeError(m)

            # Fix for https://github.com/GRIDAPPSD/GOSS-GridAPPS-D/issues/1072
            # register the onmessage from the passed listener object.  This is not
            # necessarily an ideal solution because we aren't also passing on the
            # other functions in the lifecycle, however this does pass the
            # test that was written so that listeners aren't called multiple times.
            self._router_callback.add_callback(topic, callback.on_message)
            # self._conn.set_listener(conn_id,
            #                         CallbackWrapperListener(callback.on_message, conn_id))

        _log.debug("Subscribing to {topic}".format(topic=topic))
        self._conn.subscribe(destination=topic, ack='auto', id=conn_id)

        return conn_id

    def unsubscribe(self, conn_id):
        self._conn.unsubscribe(conn_id)

    def _make_connection(self):
        if self._conn is None or not self._conn.is_connected():
            _log.debug("Creating connection")
            if not self.__token:

                # get token
                # get initial connection
                dt=datetime.now()
                replyDest = f"temp.token_resp.{self.__user__}-{dt}"

                # create token request string
                userAuthStr = f"{self.__user__}:{self.__pass__}"
                base64Str = base64.b64encode(userAuthStr.encode())

                # set up token callback
                # send request to token topic
                tokenTopic = "/topic/pnnl.goss.token.topic"

                tmpConn = Connection([(self.stomp_address, self.stomp_port)])
                if self._override_thread_fc is not None:
                    tmpConn.transport.override_threading(self._override_thread_fc)
                tmpConn.connect(self.__user__, self.__pass__, wait=True)
                
                class TokenResponseListener():
                    def __init__(self):
                        self.__token = None

                    def get_token(self):
                        return self.__token

                    def on_message(self, header, message):
                        _log.debug("Internal on message is: {} {}".format(header, message))
                        
                        self.__token = str(message)

                    def on_error(self, headers, message):
                        _log.error("ERR: {}".format(headers))
                        _log.error("OUR ERROR: {}".format(message))

                    def on_disconnect(self, header, message):
                        _log.debug("Disconnected")

                # receive token and set token variable
                # set callback
                listener = TokenResponseListener()

                # self.subscribe(replyDest, listener)
                tmpConn.subscribe('/queue/'+replyDest, 123)
                tmpConn.set_listener('token_resp', listener)
                tmpConn.send(body=base64Str, destination=tokenTopic,
                             headers={'reply-to': replyDest})
                # while token is null or for x iterations
                iter = 0
                while not self.__token and iter < 10:
                    # wait
                    self.__token = listener.get_token()
                    sleep(1)
                    iter += 1

            self._conn = Connection([(self.stomp_address, self.stomp_port)])
            if self._override_thread_fc is not None:
                self._conn.transport.override_threading(self._override_thread_fc)
            try:
                self._conn.connect(self.__token, "", wait=True)
            except TypeError as e:
                _log.error("TypeError: {e}".format(e=e))
            except NotConnectedException as e:
                _log.error("NotConnectedException: {e}".format(e=e))
            except AttributeError as e:
                _log.error("AttributeError: {e}".format(e=e))


class CallbackRouter(object):

    def __init__(self):
        self.callbacks = {}
        self._topics_callback_map = defaultdict(list)
        self._queue_callerback = Queue()
        self._thread = threading.Thread(target=self.run_callbacks)
        self._thread.daemon = True
        self._thread.start()

    def run_callbacks(self):
        _log.info("Starting thread queue")
        while True:
            cb, hdrs, msg = self._queue_callerback.get()
            try:
                msg = json.loads(msg)
            except:
                pass
                # msg = message

            for c in cb:
                c(hdrs, msg)
            sleep(0.01)

    def add_callback(self, topic, callback):
        if not topic.startswith('/topic/') and not topic.startswith('/temp-queue/'):
            topic = "/queue/{topic}".format(topic=topic)
        if callback in self._topics_callback_map[topic]:
            raise ValueError("Callbacks can only be used one time per topic")
        _log.debug("Added callbac using topic {topic}".format(topic=topic))
        self._topics_callback_map[topic].append(callback)

    def remove_callback(self, topic, callback):
        if topic in self._topics_callback_map:
            callbacks = self._topics_callback_map[topic]
            try:
                callbacks.remove(callback)
            except ValueError:
                pass

    def on_message(self, headers, message):
        destination = headers['destination']
        # _log.debug("Topic map keys are: {keys}".format(keys=self._topics_callback_map.keys()))
        if destination in self._topics_callback_map:
            self._queue_callerback.put((self._topics_callback_map[destination], headers, message))
        else:
            _log.error("INVALID DESTINATION {destination}".format(destination=destination))

    def on_error(self, header, message):
        _log.error("Error in callback router")
        _log.error(header)
        _log.error(message)



from logging import FATAL, ERROR, WARN, WARNING, INFO, DEBUG, NOTSET
import os

from . import topics as t

_nameToLevel = {
    'FATAL': FATAL,
    'ERROR': ERROR,
    'WARN': WARN,
    'WARNING': WARN,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'NOTSET': NOTSET,
}

_levelToName = {
    FATAL: 'FATAL',
    ERROR: 'ERROR',
    WARNING: 'WARN',
    WARN: 'WARN',
    INFO: 'INFO',
    DEBUG: 'DEBUG',
    NOTSET: 'NOTSET',
}

VALID_LOG_LEVELS = set(_nameToLevel.values())


def getNameToLevel(level: str) -> int:
    data = level.upper()
    return _nameToLevel.get(data, NOTSET)


class Logger:
    """
    The `Logger` class handles logging to the main gridappsd server.
    """
    def __init__(self, gaps, level=INFO):
        self._gaps = gaps
        self._level = level

    def setLevel(self, level):
        self._level = level

    def debug(self, message):
        self.log(message)

    def info(self, message):
        self.log(message, INFO)

    def error(self, message):
        self.log(message, ERROR)

    def warning(self, message):
        self.log(message, WARNING)

    def fatal(self, message):
        self.log(message, FATAL)

    def log(self, message, level=DEBUG):
        """
        Log message to the gridappsd logging api.

        :raises: AttributeError
            if the environment doesn't have GRIDAPPSD_APPLICATION_ID or GRIDAPPSD_SERVICE_ID or
            GRIDAPPSD_PROCESS_ID in the environment
        """
        process_identifier = self._gaps.get_application_id()

        if not level in VALID_LOG_LEVELS:
            raise AttributeError(f"Log level must be one of {[x for x in _levelToName.values()]}")

        if not process_identifier:
            raise AttributeError(f"Must have GRIDAPPSD_APPLICATION_ID or GRIDAPPSD_SERVICE_ID or GRIDAPPSD_PROCESS_ID "
                                 "set in os environments.")
        status = self._gaps.get_application_status()
        sim_id = self._gaps.get_simulation_id()

        if sim_id is not None:
            topic = t.simulation_log_topic(sim_id)
        else:
            topic = t.platform_log_topic()

        status_message = {
            "source": process_identifier,
            "processId": f"{sim_id}",
            "processStatus": str(status),
            "logMessage": str(message),
            "logLevel": _levelToName[level],
            "storeToDb": True
        }
        
        gridappsd_log_level = os.getenv('GRIDAPPSD_LOG_LEVEL')
        if gridappsd_log_level == None:
            gridappsd_log_level = level
        else:
            gridappsd_log_level = _nameToLevel[gridappsd_log_level]
            
        if level >= gridappsd_log_level:
            self._gaps.send(topic, status_message)




import datetime, time
from enum import Enum
from typing import Optional

from dateutil import parser
import os
try: # python2.7
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


__GRIDAPPSD_URI__ = os.environ.get("GRIDAPPSD_URI", "localhost:61613")

if not __GRIDAPPSD_URI__.startswith("tcp://"):
    __GRIDAPPSD_URI__ = "tcp://" + __GRIDAPPSD_URI__
__GRIDAPPSD_URI_PARSED__ = urlparse(__GRIDAPPSD_URI__)


def datetime_to_epoche(dt):
    return int(time.mktime(dt.timetuple()) * 1000)


def datestr_to_epoche(dt_str):
    dt = parser.parse(dt_str)
    return datetime_to_epoche(dt)


def epoche_to_datetime(epoche):
    return datetime.datetime.fromtimestamp(epoche)


def utc_timestamp():
    return datetime_to_epoche(datetime.datetime.utcnow())


def validate_gridappsd_uri():
    problems = []

    gridapspd_uri = __GRIDAPPSD_URI__
    if not gridapspd_uri.startswith("tcp://"):
        gridapspd_uri = "tcp://" + gridapspd_uri

    gridapspd_parsed_uri = urlparse(gridapspd_uri)

    if not gridapspd_parsed_uri.port:
        problems.append("Invalid port specified in URI modify environment GRIDAPPSD_URI")

    if not gridapspd_parsed_uri.hostname:
        problems.append("Invalid hostname not specified!")

    return problems


def get_gridappsd_address():
    """
    Returns the address in such a way that the response will be
    able to be passed directly to a socket and/or the stomp library.
    :return: tuple(address, port)
    """
    # New way is to specify the address in environment GRIDAPPSD_ADDRESS and use that
    # for this utility function
    address = os.environ.get("GRIDAPPSD_ADDRESS")
    if address:
        port = os.environ.get("GRIDAPPS_PORT", "61613")
        if address.startswith("tcp://"):
            address = address.replace("tcp://", "")

        return address, port

    return (__GRIDAPPSD_URI_PARSED__.hostname,
            __GRIDAPPSD_URI_PARSED__.port)


def get_gridappsd_application_id():
    """ Retrieve the application_id from the environment.

    In order to use this function an environmental variable `GRIDAPPSD_APPLICATION_ID`
    must have been set.  For docker containers this is done in the
    `gridappsd.app_registration` callback when the application is started.  If the
    environmental variable is not set an AttributeError will be raised.
    """
    app_id = os.environ.get("GRIDAPPSD_APPLICATION_ID")
    if not app_id:
        raise AttributeError("environmental variable for GRIDAPPSD_APPLICATION_ID is not set")
    return app_id


def get_gridappsd_simulation_id() -> Optional[str]:
    """ Retrieve simulation_id from environment.

    This method will return a `None` if the GRIDAPPSD_SIMULATION_ID environmental
    variable is not set.

    return: Optional[str]: simulation_id or None
    """
    simulation_id = os.environ.get("GRIDAPPSD_SIMULATION_ID", None)
    return None if simulation_id is None else str(simulation_id)


class ProcessStatusEnum(Enum):
    STARTING = "STARTING"
    STOPPING = "STOPPING"
    RUNNING = "RUNNING"
    CLOSED = "CLOSED"
    ERROR = "ERROR"
    COMPLETE = "COMPLETE"
    PAUSED = "PAUSED"
    STARTED = "STARTED"



import json
import logging
import os
import sys
import time
import stomp
from gridappsd import ApplicationController, GridAPPSD, utils

def main():
    loglevel = logging.INFO
    logging.basicConfig(stream=sys.stdout, level=loglevel,
                        format="%(asctime)s - %(name)s;%(levelname)s|%(message)s",
                        datefmt="%Y-%m-%d %H:%M:%S")
    logging.getLogger('stomp.py').setLevel(logging.ERROR)
    _log = logging.getLogger(__name__)

    problems = utils.validate_gridappsd_uri()

    if problems:
        for p in problems:
            _log.error(p)
        sys.exit(1)

    if not os.path.isfile("/appconfig"):
        _log.error("Invalid /appconfig reference...map the /appconfig to your container")
        sys.exit(1)

    config = json.loads(open("/appconfig").read())

    if "id" not in config:
        _log.error("Invalid appconfig, must have a unique id set.")
        sys.exit(1)

    os.environ['GRIDAPPSD_APPLICATION_ID'] = config['id']

    appreg = None
    gap = None
    while True:

        try:
            if gap is None:
                gap = GridAPPSD()

        except ConnectionRefusedError:  # Python 3 specific error code
            _log.debug("Retry in 10 seconds")
            gap = appreg = None
            time.sleep(10)
        except (stomp.exception.ConnectFailedException, OSError):
            _log.debug("Connect failed Retry in 10 seconds")
            gap = appreg = None
            time.sleep(10)
        except KeyboardInterrupt:
            gap = appreg = None
            _log.info("Exiting app")
            break
        else:
            if appreg is None:
                def end_app():
                    sys.exit(0)

                # app_config_minimal = {
                #     'id': 'an-app-id',
                #     'execution_path': '/home/osboxes/git/gridappsd-python/testapp.sh'
                # }
                try:
                    appreg = ApplicationController(config, gridappsd=gap)
                    appreg.register_app(end_app)
                    _log.info('Application {} registered.'.format(config['id']))
                except:
                    _log.exception("An unhandled exception occured retrying app")
                    appreg = None
                    gap = None
            else:
                if not appreg.heartbeat_valid:
                    appreg = None
                    gap = None

            time.sleep(2)





import json
import logging
import os
try:
    from queue import Queue
except ImportError:
    from Queue import Queue
import time
import select
import subprocess
import threading
import shlex
import sys
import os

from . gridappsd import GridAPPSD
from . topics import REQUEST_REGISTER_APP
from . import utils

_log = logging.getLogger(__name__)

# determine OS type
posix = False
if os.name == 'posix':
    posix = True


class Job(threading.Thread):
    def __init__(self, args, out=sys.stdout, err=sys.stderr):
        threading.Thread.__init__(self)
        _log.debug("Creating job")
        self.running = False
        self._args = args
        self._out = out
        self._err = err

    def shutdown(self):
        self.running = False

    def run(self):
        try:
            self.running = True
            os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'RUNNING'

            p = subprocess.Popen(args=self._args,
                                 shell=False,
                                 stdout=self._out,
                                 stderr=self._err)

            # Loop while process is executing
            while p.poll() is None and self.running:
                os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'RUNNING'
                time.sleep(1)

        except Exception as e:
            os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'ERROR'
            _log.error(repr(e))
        else:
            os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPED'


class ApplicationController(object):

    def __init__(self, config, gridappsd=None, heatbeat_period=10):
        if not isinstance(config, dict):
            raise ValueError("Config should be dictionary")
        if not isinstance(gridappsd, GridAPPSD):
            raise ValueError("Invalid gridappsd instance passed.")

        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPED'
        self._configDict = config.copy()
        self._validate_config()
        self._gapd = gridappsd
        self._shutting_down = False
        self._heartbeat_thread = None
        self._heartbeat_period = heatbeat_period
        self._heartbeat_should_stop = False
        self._application_id = None
        self._heartbeat_topic = None
        self._start_control_topic = None
        self._stop_control_topic = None
        self._status_control_topic = None
        self._thread = None
        self._jobs = []
        self._process_start_time = None
        self._process_end_time = None
        self._process_is_running = False
        self._process_has_started = False
        self._end_callback = None
        self._print_queue = Queue()
        self._heartbeat_thread = None
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPED'

        if "type" not in self._configDict or self._configDict['type'] != 'REMOTE':
            _log.warning('Setting type to REMOTE you can remove this error by putting '
                         '"type": "REMOTE" in the app config file.')
            self._configDict['type'] = 'REMOTE'

    def _validate_config(self):
        required = ['id', 'execution_path']
        missing = [x for x in required if x not in self._configDict]

        if missing:
            raise ValueError("Missing {} in config object.".format(missing))

    @property
    def application_id(self):
        return self._application_id

    @property
    def heartbeat_valid(self):
        return self._heartbeat_thread is not None

    def register_app(self, end_callback):
        print("Sending {}\n\tto {}".format(self._configDict,
                                           REQUEST_REGISTER_APP))
        self._gapd.get_logger().debug("Started App Registration")

        response = self._gapd.get_response(REQUEST_REGISTER_APP,
                                           self._configDict,
                                           60)
        if 'message' in response:
            _log.error("An error regisering the application occured")
            _log.error(response.get('message'))
            raise ValueError(response.get('message'))
        self._application_id = response.get('applicationId')
        self._heartbeat_topic = response.get('heartbeatTopic')
        self._heartbeat_period = response.get('heartbeatPeriod', 10)
        self._start_control_topic = response.get('startControlTopic')
        self._stop_control_topic = response.get('stopControlTopic')

        os.environ["GRIDAPPSD_APPLICATION_ID"] = self._application_id
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPED'

        self._gapd.subscribe(self._stop_control_topic, self.__handle_stop)
        self._gapd.subscribe(self._start_control_topic, self.__handle_start)
        self._end_callback = end_callback

        # TODO assuming good response start the heartbeat
        self._heartbeat_thread = threading.Thread(target=self.__start_heartbeat,
                                                  args=[self.__heartbeat_error])
        self._heartbeat_thread.daemon = True
        self._heartbeat_thread.start()
        self._gapd.get_logger().debug("Heartbeat registereed for application {}".format(
            utils.get_gridappsd_application_id()))

    def __heartbeat_error(self):
        self._heartbeat_thread = None

    def __start_heartbeat(self, error_callback):
        starttime = time.time()

        try:
            while True:
                self._print_queue.put("Sending heartbeat for {}".format(self._application_id))
                # print("Seanding heartbeat {} {}".format(self._heartbeat_topic, self._application_id))
                # print("Heartbeat period: {}".format(self._heartbeat_period))
                self._gapd.send(self._heartbeat_topic, self._application_id)
                time.sleep(self._heartbeat_period - ((time.time() - starttime) % self._heartbeat_period))
        except:
            error_callback()

    def __print_from_queue(self):
        while True:
            buff = self._print_queue.get(block=True)
            print(buff)

    def __handle_start(self, headers, message):
        _log.debug("Handling start")
        if isinstance(message, str):
            obj = json.loads(message)
        else:
            obj = message
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STARTING'
        self._gapd.get_logger().debug("Handling Start: {}\ndict:\n{}".format(headers, obj))

        if 'command' not in obj:
            # Send log to gridappsd
            _log.error("Invalid message sent on start app.")
        else:
            _log.debug("CWD IS: {}".format(os.getcwd()))
            args = shlex.split(obj['command'])
            job = Job(args)
            job.daemon = True
            job.start()
            self._gapd.get_logger().debug("Job Started: {}".format(job.running))

    def __handle_stop(self, headers, message):
        print("Handling Stop: {} {}".format(headers, message))
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPING'
        if self._thread:
            self._thread.join()
        if self._end_callback is not None:
            self._end_callback()
        os.environ['GRIDAPPSD_APPLICATION_STATUS'] = 'STOPPED'

    def shutdown(self):
        self._shutting_down = True



import json
from gridappsd import topics
from typing import Optional



class Query:
    """ Class to create and execute request to query timeseries data
    """
    
    def __init__(self, measurement):
        self.queryMeasurement = measurement
        self.queryFilter = []
        self.selectCriteria = []
        self.key = None

    def select(self, *keys):
        """ Defines what fields should be returned from the query.
        
        If this function is not called or is called without argument then 
        all the fields are returned.
        
        :param keys: list of fields to be returned 
        """
        for key in keys:
            self.selectCriteria.append(key)
        return self

    def first(self, n=Optional[int]):
        """ Method to add request to return first or oldest data to the query.
        
        When n is specified, query will return first or oldest 'n' rows.
        
        :param n:  
        """
        if n is not None:
            self.first = n;
        return self

    def last(self, n=Optional[int]):
        """ Method to add request to return last or latest data to the query.
        
        When n is specified, query will return last or latest 'n' rows.
        
        :param n:  
        """
        if n is not None:
            self.last = n;
        return self
    
    def ge(self, value):
        """ Method to add 'value greater than or equal to' filter to a key.
        
        :param value:  
        """
        if self.key is None:
            raise ValueError("Key is not specified. Call where_key first.")
        obj = {"key":self.key,"ge":value}
        self.queryFilter.append(obj)
        return self
    
    def le(self, value):
        """ Method to add 'value less than or equal to' filter to a key.
        
        :param value:  
        """
        if self.key is None:
            raise ValueError("Key is not specified. Call where_key first.")
        obj = {"key":self.key,"le":value}
        self.queryFilter.append(obj)
        return self
    
    def eq(self, value):
        """ Method to add 'value equal to' filter to a key.
        
        :param value:  
        """
        if self.key is None:
            raise ValueError("Key is not specified. Call where_key first.")
        obj = {"key":self.key,"eq":value}
        self.queryFilter.append(obj)
        return self
    
    def between(self, value1, value2):
        """ Method to add 'value between' value1 and value2 filter to a key.
        
        :param value1: defines 'greater than equal to' filter for key's value
        :param value2: defines 'less than equal to' filter for key's value
        """
        if self.key is None:
            raise ValueError("Key is not specified. Call where_key first.")
        obj = {"key":self.key,"ge":value1}
        self.queryFilter.append(obj)
        obj = {"key":self.key,"le":value2}
        self.queryFilter.append(obj)
        return self
    
    def where_key(self, key):
        self.key = key
        return self

    def execute(self, gridappsd_obj):
        del self.key 
        response = gridappsd_obj.get_response(topics.REQUEST_TIMESERIES_DATA, json.dumps(self.__dict__))
        return response


# -------------------------------------------------------------------------------
# Copyright (c) 2018, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import inspect
import json
import logging
from datetime import datetime
from logging import DEBUG, INFO, WARNING, FATAL, WARN
import time

from . import GOSS, utils
from . import topics as t
from . import ProcessStatusEnum
from .houses import Houses
from .loghandler import Logger, VALID_LOG_LEVELS, getNameToLevel

# from . configuration_types import ConfigurationType

_log = logging.getLogger(inspect.getmodulename(__file__))

POWERGRID_MODEL = "powergridmodel"


class InvalidSimulationIdError(Exception):
    pass


class GridAPPSD(GOSS):
    """ The main :class:`GridAPPSD` interface for connecting to a GridAPPSD instance
    """
    # TODO Get the caller from the traceback/inspect module.
    def __init__(self, simulation_id=None,
                 address=None, **kwargs):

        if address is None:
            address = utils.get_gridappsd_address()

        if 'stomp_address' in kwargs and 'stomp_port' in kwargs:
            address = (kwargs.pop('stomp_address'), kwargs.pop('stomp_port'))
        elif 'stomp_address' in kwargs and not 'stomp_port' in kwargs or \
             'stomp_port' in kwargs and not 'stomp_address' in kwargs:
            raise ValueError("If stomp_address is specified the so should stomp_port")

        super(GridAPPSD, self).__init__(
            stomp_address=address[0],
            stomp_port=address[1],
            **kwargs)
        self._houses = Houses(self)
        self._simulation_log_topic = None
        self._simulation_id = None
        # Transfer simulation_id from environment if its not passed
        # through the constructor.
        if simulation_id is None:
            simulation_id = utils.get_gridappsd_simulation_id()
        self._simulation_id = simulation_id
        self._process_status = ProcessStatusEnum("STARTING")
        if self._simulation_id:
            self._simulation_log_topic = t.simulation_log_topic(self._simulation_id)

    def get_logger(self) -> Logger:
        """
        Return a log instance for interacting with the log within the gridappsd platform

        :return: :class:`gridappsd.loghandler.Logger`
        """
        return Logger(self)

    def get_houses(self) -> Houses:
        return self._houses
    
    def get_simulation_id(self):
        """ 
        Return the simulation_id given to the GridAPPSD instance
        
        :return:
        """
        if not self._simulation_id:
            self._simulation_id = utils.get_gridappsd_simulation_id()
            if self._simulation_id:
                self._simulation_log_topic = t.simulation_log_topic(self._simulation_id)
        return self._simulation_id
    
    def set_application_status(self, status):
        """
        Set the application status.
        :param status:
        """
        try:
            self._process_status = ProcessStatusEnum(status)
        except ValueError:
            self.get_logger().warning("Unsuccessful change of application status."
                            + f"Valid statuses are {ProcessStatusEnum.__members__}.")

    def set_service_status(self, status):
        """
        Set the service status.
        :param status:
        """
        try:
            self._process_status = ProcessStatusEnum(status)
        except ValueError:
            self.get_logger().warning("Unsuccessful change of service status."
                                      + f"Valid statuses are {ProcessStatusEnum.__members__}.")

    def set_simulation_id(self, simulation_id):
        if simulation_id is None:
            self.get_logger().warning("None value for simulation_id set.")
        else:
            simulation_id = str(simulation_id)
            self._simulation_id = simulation_id
            self._simulation_log_topic = t.simulation_log_topic(self._simulation_id)

    def get_application_status(self):
        """
        Return the application status
        :return:
        """
        return self._process_status.value

    def get_application_id(self):
        return utils.get_gridappsd_application_id()
    
    def get_service_status(self):
        """
        Return the service status
        :return:
        """
        return self._process_status.value
    
    def query_object_types(self, model_id=None):
        """ Allows the caller to query the different object types.
                
        :param model_id:
        :return:
        """
        args = {}
        if model_id:
            args["modelId"] = model_id
        payload = self._build_query_payload("QUERY_OBJECT_TYPES", **args)
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)

    def query_model_names(self, model_id=None):
        args = {}
        if model_id is not None:
            args["modelId"] = model_id
        payload = self._build_query_payload("QUERY_MODEL_NAMES", **args)
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)

    def query_model_info(self):
        payload = self._build_query_payload("QUERY_MODEL_INFO")
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)
    
    def query_model(self, model_id=None, object_type=None, object_id=None, response_format='JSON'):
        args = {}
        if model_id is not None:
            args["modelId"] = model_id
        if object_type is not None:
            args["objectType"] = object_type
        if object_id is not None:
            args["objectId"] = object_id   
        if response_format is not None:
            args["resultFormat"] = response_format   
        payload = self._build_query_payload("QUERY_MODEL", **args)
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)

    def query_object(self, object_id, model_id=None):
        if not object_id:
            raise ValueError("Invalid object_id specified.")
        args = dict(objectId=object_id)
        if model_id is not None:
            args["modelId"] = model_id
        payload = self._build_query_payload("QUERY_OBJECT", **args)
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)
    
    def query_object_dictionary(self, model_id, object_type=None, object_id=None):
        if not model_id:
            raise ValueError("model_id is not specified.")
        if not object_id and not object_type:
            raise ValueError("No obejct_id or object_type specified.")
        args={}
        args["modelId"] = model_id
        if object_id is not None:
            args["objectId"] = object_id
        if object_type is not None:
            args["objectType"] = object_type
        payload = self._build_query_payload("QUERY_OBJECT_DICT", **args)
        return self.get_response(t.REQUEST_POWERGRID_DATA, payload, timeout=30)

    def query_data(self, query, database_type=POWERGRID_MODEL, timeout=30):
        request_type = None
        if database_type == POWERGRID_MODEL:
            request_type = 'QUERY'
        else:
            raise ValueError("Only supported {} currently".format(POWERGRID_MODEL))

        payload = self._build_query_payload(request_type, queryString=query)
        # Do this so we can eventually support other db through this mechanism.
        request_topic = '.'.join((t.REQUEST_DATA, database_type))
        return self.get_response(request_topic, json.dumps(payload), timeout=timeout)

    def get_platform_status(self, applications=True, services=True, appInstances=True, serviceInstances=True):
        _log.debug("Retrieving platform status from GridAPPSD")
        msg = dict(appInstances=appInstances, applications=applications, services=services,
                   serviceInstances=serviceInstances)
        return self.get_response(t.REQUEST_PLATFORM_STATUS, json.dumps(msg), timeout=30)

    def send_simulation_status(self, status, message, log_level=INFO):
        _log.debug("SEND SIM STATUS: {} message: {}".format(status, message))
        # Transform log level into something for use if we can.
        if isinstance(log_level, str):
            log_level = getNameToLevel(log_level)
        if log_level not in VALID_LOG_LEVELS:
            raise InvalidSimulationIdError("Invalid log_level specified!")
        if not self._simulation_log_topic:
            raise InvalidSimulationIdError()
        gad_log = self.get_logger()
        self.set_service_status(status)
        gad_log.log(message, log_level)

    def send_status(self, status, topic, log_level=INFO):
        self.set_application_status(status)
        status_message = self.build_message_json(self.get_application_status(),
            "", log_level)
        self.send(topic, status_message)

    def build_message_json(self, status, message, log_level):
        t_now = datetime.utcnow()
        status_message = {
            "source": utils.get_gridappsd_application_id(),
            "processId": f"{self._simulation_id}",
            "timestamp": int(time.mktime(t_now.timetuple()))*1000,
            "procesStatus": status,
            "logMessage": str(message),
            "logLevel": logging.getLevelName(log_level),
            "storeToDb": True
        }
        data = json.dumps(status_message)

        return data

    def _build_query_payload(self, request_type, response_format='JSON', **kwargs):
        d = dict(requestType=request_type, resultFormat=response_format)
        d.update(**kwargs)
        return d






# -------------------------------------------------------------------------------
# Copyright (c) 2018, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import logging
import sys
import argparse
from argparse import ArgumentParser
from time import sleep
import yaml

from gridappsd import GridAPPSD

assert sys.version_info >= (3, 6), "Minimum version is python 3.6"

logging.basicConfig(stream=sys.stdout, level=logging.INFO,
                    format="'%(asctime)s: %(name)-20s - %(levelname)-6s - %(message)s")

logging.getLogger('stomp.py').setLevel(logging.WARNING)
_log = logging.getLogger("gridappsd.__main__")


if __name__ == '__main__':

    parser = ArgumentParser()

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "--run-simulation", type=argparse.FileType('r'),
                       help="Start running a simulation from a passed simulation file.")

    opts = parser.parse_args()

    if opts.run_simulation:
        # Example of how to pause simulation
        def next_timestep(simulation, timestep):
            simulation.pause()
            sleep(1)
            simulation.resume()

        gappsd = GridAPPSD()
        run_args = yaml.safe_load(opts.run_simulation)

        # if wanting to use the above next_timestep function use this
        # instead of the one below.
        # simulation = gappsd.run_simulation(run_args, next_timestep)
        simulation = gappsd.run_simulation(run_args)
        simulation.simulation_main_loop()



#!/usr/bin/env python3

import contextlib
import logging
import os
import random
import re
import shutil
import tarfile
import threading
import urllib.request
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from pprint import pformat
from subprocess import PIPE
from typing import Optional, Union

import stomp
import time
from docker.models.containers import Container

from gridappsd import GridAPPSD
from gridappsd.goss import GRIDAPPSD_ENV_ENUM

logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
logging.getLogger("docker.auth").setLevel(logging.INFO)
logging.getLogger("docker.utils").setLevel(logging.INFO)

_log = logging.getLogger("gridappsd.docker_handler")


try:
    import docker
    HAS_DOCKER = True
except ImportError:
    _log.warning("Docker api not loaded. pip install docker to install as package.")
    HAS_DOCKER = False

if HAS_DOCKER:

    # The following variable is used for creating a volume for the gridappsd container
    # to utilize.  It allows the ability to use multiple containers to run tests
    # along side each other.
    GRIDAPPSD_CONFIG_VOLUME_NAME = f"gridappsd_config_{random.randint(1,100)}"

    # This named container will be used to hold configuration/folders so that other containers
    # that start can use them.  To use add "volume_from": [CONFIGURATION_CONTAINER_NAME] and
    # the mount point within the container will also be within the service container.
    CONFIGURATION_CONTAINER_NAME = "testconfig"

    def expand_all(user_path):
        return os.path.expandvars(os.path.expanduser(user_path))

    __TMP_ROOT__ = "/tmp/assets"
    if Path(__TMP_ROOT__).exists():
        shutil.rmtree(__TMP_ROOT__, ignore_errors=True)
    os.makedirs(__TMP_ROOT__)

    # This path needs to be the path to the repo where configuration files are located.
    GRIDAPPSD_CONF_DIR = Path(__file__).resolve().parent.parent.joinpath("gridappsd/conf")

    assert Path(GRIDAPPSD_CONF_DIR).joinpath("entrypoint.sh").exists()
    assert Path(GRIDAPPSD_CONF_DIR).joinpath("run-gridappsd.sh").exists()

    GRIDAPPSD_DATA_REPO = str(Path(__TMP_ROOT__).joinpath("mysql").resolve())

    os.makedirs(GRIDAPPSD_DATA_REPO, exist_ok=True)
    if not Path(GRIDAPPSD_DATA_REPO).is_dir():
        raise AttributeError(f"Invalid GRIDAPPSD_DATA_REPO couldn't make or doesn't exist {GRIDAPPSD_DATA_REPO}")

    MYSQL_SCHEMA_INIT_DIR = f'{GRIDAPPSD_DATA_REPO}/docker-entrypoint-initdb.d'


    def mysql_setup():
        """
        Downloads gridappsd_mysql_dump.sql into the MYSQL_SCHEMA_INIT_DIR init directory.
        This will then be mounted into the mysql container.
        """
        # Downlaod mysql file
        _log.debug("Downloading mysql data file from Bootstrap repository")
        mysql_file = f'{MYSQL_SCHEMA_INIT_DIR}/gridappsd_mysql_dump.sql'
        if os.path.isdir(mysql_file):
            raise RuntimeError(f"mysql datafile is directory, delete {mysql_file} using sudo rm -rf {mysql_file}")
        if not os.path.isdir(MYSQL_SCHEMA_INIT_DIR):
            os.makedirs(MYSQL_SCHEMA_INIT_DIR, 0o0775, exist_ok=True)
        urllib.request.urlretrieve(
            'https://raw.githubusercontent.com/GRIDAPPSD/Bootstrap/master/gridappsd_mysql_dump.sql',
            filename=mysql_file)

        # Modify the mysql file to allow connections from gridappsd container
        with open(mysql_file, "r") as sources:
            lines = sources.readlines()
        with open(mysql_file, "w") as sources:
            for line in lines:
                sources.write(re.sub(r'localhost', '%', line))
        assert Path(mysql_file).exists()

    # Use the environmental variable if specified otherwise use the develop tag.
    DEFAULT_GRIDAPPSD_TAG = os.environ.get('GRIDAPPSD_TAG_ENV', 'develop')

    # Network to connect all of the containers up to by default.
    NETWORK = "test_my_network"

    __TPL_DEPENDENCY_CONFIG__ = {
        "influxdb": {
            "start": True,
            "image": "gridappsd/influxdb:{{DEFAULT_GRIDAPPSD_TAG}}",
            "pull": True,
            "ports": {"8086/tcp": 8086},
            "environment": {"INFLUXDB_DB": "proven"}
        },
        "redis": {
            "start": True,
            "image": "redis:3.2.11-alpine",
            "pull": True,
            "ports": {"6379/tcp": 6379},
            "environment": [],
            "entrypoint": "redis-server --appendonly yes",
        },
        "blazegraph": {
            "start": True,
            "image": "gridappsd/blazegraph:{{DEFAULT_GRIDAPPSD_TAG}}",
            "pull": True,
            "ports": {"8080/tcp": 8889},
            "environment": []
        },
        "mysql": {
            "start": True,
            "image": "mysql/mysql-server:5.7",
            "pull": True,
            "ports": {"3306/tcp": 3306},
            "environment": {
                "MYSQL_RANDOM_ROOT_PASSWORD": "yes",
                "MYSQL_PORT": "3306"
            },
            # "volumes": {
            #     "/home/gridappsd/test-assets": {"bind": "/whatthehell", "mode": "rw"}
            #     #"test-assets": {"bind": "/whatthehell", "mode": "rw"}
            #     #data_dir + "/dumps/": {"bind": "/docker-entrypoint-initdb.d/", "mode": "ro"}
            # },
            # Our own so we can create
            "volumes_required": [
                dict(name="mysql_config",
                     local_path=MYSQL_SCHEMA_INIT_DIR,
                     container_path="/docker-entrypoint-initdb.d")
            ],
            "volumes_from": [
                "mysql_config"
            ],
            "onsetupfn": mysql_setup
        },
        "proven": {
            "start": True,
            "image": "gridappsd/proven:{{DEFAULT_GRIDAPPSD_TAG}}",
            "pull": True,
            "ports": {"8080/tcp": 18080},
            "environment": {
                "PROVEN_SERVICES_PORT": "18080",
                "PROVEN_SWAGGER_HOST_PORT": "localhost:18080",
                "PROVEN_USE_IDB": "true",
                "PROVEN_IDB_URL": "http://influxdb:8086",
                "PROVEN_IDB_DB": "proven",
                "PROVEN_IDB_RP": "autogen",
                "PROVEN_IDB_USERNAME": "root",
                "PROVEN_IDB_PASSWORD": "root",
                "PROVEN_T3DIR": "/proven"},
            "links": {"influxdb": "influxdb"}
        }
    }

    __TPL_GRIDAPPSD_CONFIG__ = {
        "gridappsd": {
            "start": True,
            "image": "gridappsd/gridappsd:{{DEFAULT_GRIDAPPSD_TAG}}",
            "pull": True,
            "ports": {"61613/tcp": 61613, "61614/tcp": 61614, "61616/tcp": 61616},
            "environment": {
                "PATH": "/gridappsd/bin:/gridappsd/lib:/gridappsd/services/fncsgossbridge/service:/usr/local/bin:/usr/local/sbin:/usr/sbin:/usr/bin:/sbin:/bin",
                "DEBUG": 1,
                "START": 1
            },
            "links": {"mysql": "mysql",
                      "influxdb": "influxdb",
                      "blazegraph": "blazegraph",
                      "proven": "proven",
                      "redis": "redis"},
            "volumes_required": [
                dict(name=GRIDAPPSD_CONFIG_VOLUME_NAME,
                     local_path=GRIDAPPSD_CONF_DIR,
                     container_path="/startup/conf")
            ],
            "volumes_from": [
                GRIDAPPSD_CONFIG_VOLUME_NAME
            ],
            "entrypoint": "/startup/conf/entrypoint.sh",
            "command": "/startup/conf/entrypoint.sh"
        }
    }

    def __update_template_data__(data, update_dict):
        data_cpy = deepcopy(data)
        for k, v in data_cpy.items():
            for u, p in update_dict.items():
                v['image'] = v['image'].replace(u, p)

        return data_cpy


    __replace_dict__ = {"{{DEFAULT_GRIDAPPSD_TAG}}": DEFAULT_GRIDAPPSD_TAG}
    DEFAULT_DOCKER_DEPENDENCY_CONFIG = __update_template_data__(__TPL_DEPENDENCY_CONFIG__, __replace_dict__)
    DEFAULT_GRIDAPPSD_DOCKER_CONFIG = __update_template_data__(__TPL_GRIDAPPSD_CONFIG__, __replace_dict__)

    def update_gridappsd_tag(new_gridappsd_tag):
        """
        Update the default tag used within the dependency and gridappsd containers to be
        what is specified in the new gridappsd_tag variable
        """
        global DEFAULT_GRIDAPPSD_TAG

        DEFAULT_GRIDAPPSD_TAG = new_gridappsd_tag
        _log.info(f"Updated gridappsd docker tag {DEFAULT_GRIDAPPSD_TAG} ")
        __replace_dict__.update({"{{DEFAULT_GRIDAPPSD_TAG}}": DEFAULT_GRIDAPPSD_TAG})
        DEFAULT_DOCKER_DEPENDENCY_CONFIG.update(__update_template_data__(__TPL_DEPENDENCY_CONFIG__, __replace_dict__))
        DEFAULT_GRIDAPPSD_DOCKER_CONFIG.update(__update_template_data__(__TPL_GRIDAPPSD_CONFIG__, __replace_dict__))


    class Containers:
        """
        This class allows the creation/management of containers created by the gridappsd
        docker process.
        """
        __client__ = docker.from_env()

        def __init__(self, container_def):
            self._container_def = container_def

        @property
        def container_def(self):
            return self._container_def

        @staticmethod
        def remove_container(name):
            try:
                container = Containers.__client__.containers.get(name)
                container.kill()
            except docker.errors.NotFound:
                _log.debug(f"container {name} not found so couldn't remove.")

        @staticmethod
        def container_list(ignore_list: Optional[Union[str, list]] = "gridappsd_dev"):
            """
            Provides a wrapper around the listing of docker containers.  This function was
            brought about when running from within a docker container.

            Currently the docker container that is run using docker-compose up from the
            gridappsd-dev-environment is specified as gridappsd_dev, however this can change
            and could potentially be extended to multiple named containers.

            @param: ignore_list:
                optional container names to not stop :: A string or list of strings
            """

            if ignore_list is None:
                ignore_list = []
            elif isinstance(ignore_list, str):
                ignore_list = [ignore_list]
            containers = []
            for cont in Containers.__client__.containers.list():
                if cont.name not in ignore_list:
                    containers.append(cont)
            return containers

        @staticmethod
        def reset_all_containers(ignore_list: Optional[Union[str, list]] = "gridappsd_dev"):
            """
            Provides a wrapper around the resetting of all docker containers.  This function was
            brought about when running from within a docker container.

            Currently the docker container that is run using docker-compose up from the
            gridappsd-dev-environment is specified as gridappsd_dev, however this can change
            and could potentially be extended to multiple named containers.

            @param: ignore_list:
                optional container names to not stop :: A string or list of strings
            """
            if ignore_list is None:
                ignore_list = []
            elif isinstance(ignore_list, str):
                ignore_list = [ignore_list]

            for cont in Containers.__client__.containers.list():
                if cont.name not in ignore_list:
                    cont.kill()

        @staticmethod
        def check_required_running(self, config):
            my_config = deepcopy(config)

            for c in Containers.__client__.containers.list():
                my_config.pop(c.name, None)

            assert not my_config, f"The required containers were not satisfied missing {list(my_config.keys())}"

        @staticmethod
        def create_volume_container(name, volume_name,
                                    mount_in_container_at, restart_if_exists: bool = False,
                                    mode="rw"):

            _log.info(f"Creating container {name} and mounting volume {volume_name} "
                      f"at {mount_in_container_at} in container {name}")
            client = docker.from_env()

            should_create = False
            try:
                cont = Containers.__client__.containers.get(name)
            except docker.errors.NotFound:
                # start container
                should_create = True
            else:
                if restart_if_exists:
                    should_create = True
                    cont.stop()
                else:
                    return cont

            if should_create:
                kwargs = {}
                kwargs['image'] = "alpine"
                # Only name the containers if remove is on
                kwargs['remove'] = True
                kwargs['name'] = name
                kwargs['detach'] = True
                kwargs['volumes'] = {
                    volume_name: {"bind": mount_in_container_at, "mode": mode}
                }
                # keep the container running
                kwargs['command'] = "tail -f /dev/null"
                cont = client.containers.run(**kwargs)
                _log.info(f"New container for volume created {cont.name}")
                # print(f"New container is: {container.name}")
                return cont

        @staticmethod
        def create_get_network(name: str) -> docker.models.networks.Network:
            try:
                network = Containers.__client__.networks.get(name)
            except docker.errors.NotFound:
                network = Containers.__client__.networks.create(name, driver="bridge")

            return network

        @staticmethod
        def copy_to(src: str, dst: str):
            """
            Copy a local directory onto a destination

            .. note::

                Make sure the dst (destination) is in the form container:directory

            @param: src: The local directory to copy from the host
            @param: dst: The container:destination to copy the src to.
            """
            _log.debug(f"copying folder from: {src} to {dst}")
            src = str(src)
            # python3.6 can't combine PosixPath and str
            assert os.path.exists(src), f"{src} does not exist!"
            client = docker.from_env()
            name, dst = dst.split(':')
            container = client.containers.get(name)
            assert container is not None
            cwd = os.getcwd()
            os.chdir(os.path.dirname(src))
            srcname = os.path.basename(src)
            tarfilename = src + '.tar'
            tar = tarfile.open(tarfilename, mode='w')
            try:
                tar.add(srcname)
            finally:
                tar.close()

            data = open(tarfilename, 'rb').read()
            resp = container.put_archive(os.path.dirname(dst), data)
            os.chdir(cwd)
            _log.debug(f"Response from put_archive {resp}")

        def start(self):
            _log.info("Starting containers")
            # Containers.create_volume_container(CONFIGURATION_CONTAINER_NAME, "testconfig", "/testconfig")
            #pprint(DEFAULT_GRIDAPPSD_DOCKER_CONFIG)
            client = docker.from_env()
            # print(f"Docker client version: {client.version()}")
            for service, value in self._container_def.items():
                _log.info(f"Pulling {service} image")
                _log.info(f"Pulling {service} : {self._container_def[service]['image']}")
                client.images.pull(self._container_def[service]['image'])
                try:
                    container = client.containers.get(service)
                    self._container_def[service]['containerid'] = container.id
                except docker.errors.NotFound:
                    _log.debug(f"Couldn't find {service} so continuing on.")

                # Provide a way to dynamically create things that the container will need
                # on the host system.  This is important if we want to create a volume before
                # starting the container up.
                if value.get('onsetupfn'):
                    value.get('onsetupfn')()

                if self._container_def[service].get("volumes_required"):
                    for vr in self._container_def[service].get("volumes_required"):
                        _log.info(f"Creating volume for {service}: name={vr['name']}, "
                                  f"volume_name={vr['name']}, container_path={vr['container_path']}")
                        Containers.create_volume_container(
                            name=vr["name"],
                            volume_name=vr["name"],
                            mount_in_container_at=vr["container_path"],
                            restart_if_exists=True
                            )
                        time.sleep(20)
                        if vr.get("local_path"):
                            _log.debug(f"contents of local path({vr.get('local_path')}):\n\t{os.listdir(vr.get('local_path'))}")
                            _log.info(f"Copy to mounted volume for {service}: "
                                      f"local_path={vr['local_path']}, container_path={vr['container_path']}")
                            Containers.copy_to(vr["local_path"], f'{vr["name"]}:{vr["container_path"]}')
            network = Containers.create_get_network(NETWORK)
            for service, value in self._container_def.items():
                if self._container_def[service]['start']:
                    _log.debug(f"Starting {service} : {self._container_def[service]['image']}")
                    kwargs = {}
                    kwargs['image'] = self._container_def[service]['image']
                    # Only name the containers if remove is on
                    kwargs['remove'] = True
                    kwargs['name'] = service
                    kwargs['detach'] = True
                    kwargs['entrypoint'] = value.get('entrypoint')
                    if self._container_def[service].get('entrypoint'):
                        kwargs['entrypoint'] = value['entrypoint']
                    if self._container_def[service].get('environment'):
                        kwargs['environment'] = value['environment']
                    if self._container_def[service].get('ports'):
                        kwargs['ports'] = value['ports']
                    if self._container_def[service].get('volumes'):
                        kwargs['volumes'] = value['volumes']
                    if self._container_def[service].get('entrypoint'):
                        kwargs['entrypoint'] = value['entrypoint']
                    # if self._container_def[service].get('links'):
                    #     kwargs['links'] = value['links']
                    if self._container_def[service].get('volumes_from'):
                        kwargs['volumes_from'] = value['volumes_from']
                    #for k, v in kwargs.items():
                    #    print(f"k->{k}, v->{v}")
                    _log.debug("Starting container with the following args:")
                    _log.debug(f"{pformat(kwargs)}")
                    launched_container = None
                    try:
                        container = client.containers.get(service)
                        _log.debug("Found existing container")
                    except docker.errors.NotFound:
                        container = client.containers.run(**kwargs)
                        _log.debug("Started new container")
                        network.connect(container.id)
                    self._container_def[service]['containerid'] = container.id
            _log.debug(f"Current containers are: {[x.name for x in client.containers.list()]}")

        def wait_for_log_pattern(self, container, pattern, timeout=60):
            assert self._container_def.get(container), f"Container {container} is not in definition."
            client = docker.from_env()
            container = client.containers.get(self._container_def.get(container)['containerid'])
            until = datetime.now() + timedelta(seconds=timeout)
            for p in container.logs(stream=True, until=until):
                _log.info(f"HANDLER: {p.decode('utf-8')}")
                if pattern in p.decode('utf-8'):
                    print(f"Found pattern {pattern}")
                    return
            raise TimeoutError(f"Pattern {pattern} was not found in logs of container {container} within {timeout}s")

        def wait_for_http_ok(self, url, timeout=30):
            import requests
            results = None
            test_count = 0

            while results is None or not results.ok:
                test_count += 1
                if test_count > timeout:
                    raise TimeoutError(f"Could not reach {url} within allotted timeout {timeout}s")
                results = requests.get(url)
                time.sleep(1)

            print(f"Found url {url} within timeout {timeout}")

        def stop(self):
            client = docker.from_env()
            for service, value in self._container_def.items():
                if value.get('containerid'):
                    try:
                        cnt = client.containers.get(value.get('containerid'))
                        cnt.kill()
                        time.sleep(2)

                        if "volumes_required" in value:
                            # Loop over the volumes that are required for each image and
                            # remove the volume.
                            for volume_spec in value["volumes_required"]:
                                Containers.remove_container(volume_spec['name'])
                                time.sleep(2)

                        # client.containers.get(value.get('containerid')).kill() # value.get('name')).kill()
                    except docker.errors.NotFound as ex:
                        _log.error(f"Volume {value.get('containerid')} was not found.")
                        _log.exception(ex)


    threads = []

    def stream_container_log_to_file(container_name: str, logfile: str):
        client = docker.from_env()
        container = client.containers.list(filters=dict(name=container_name))[0]

        print(container)

        def log_output():
            nonlocal container, logfile
            print(f"Starting to write to file {logfile}.")
            with open(logfile, 'wb') as fp:
                print(f"openfile")
                for p in container.logs(stream=True, stderr=PIPE, stdout=PIPE):
                    fp.write(p)
                    fp.flush()

        threading.Thread(target=log_output, daemon=True).start()


    @contextlib.contextmanager
    def run_containers(config, stop_after=True) -> Containers:
        containers = Containers(config)

        containers.start()
        try:
            yield containers
        finally:
            if stop_after:
                containers.stop()


    @contextlib.contextmanager
    def run_dependency_containers(stop_after=False) -> Containers:

        containers = Containers(DEFAULT_DOCKER_DEPENDENCY_CONFIG)

        containers.start()
        try:
            yield containers
        finally:
            if stop_after:
                containers.stop()


    @contextlib.contextmanager
    def run_gridappsd_container(stop_after=True, rebuild_if_present=False) -> Container:
        """ A contextmanager that uses """

        parent_container = get_docker_in_docker()

        client = docker.from_env()
        # if there is a parent_container then we need to make sure that it is connected
        # to the same network as our systems.  If not then we need to modify the network
        # to include the parent container
        if parent_container:
            env = DEFAULT_GRIDAPPSD_DOCKER_CONFIG['gridappsd'].get('environment')
            if env is None:
                env = {}
            env[GRIDAPPSD_ENV_ENUM.GRIDAPPSD_ADDRESS.name] = 'gridappsd'
            env[GRIDAPPSD_ENV_ENUM.GRIDAPPSD_USER.name] = 'test_app_user'
            env[GRIDAPPSD_ENV_ENUM.GRIDAPPSD_PASSWORD.name] = '4Test'
            DEFAULT_GRIDAPPSD_DOCKER_CONFIG['gridappsd']['environment'] = env

            _log.debug(f"Running inside a container environment: {parent_container.name}")
            network = client.networks.get(NETWORK)
            has_it = False
            for x in network.containers:
                if x.name == parent_container.name:
                    has_it = True
                    _log.debug(f"parent_container {parent_container.name} is connected to the network.")
                    break
            if not has_it:
                _log.debug(f"Connecting new container to the network: {parent_container.name}")
                network.connect(parent_container)
        else:
            _log.debug("Not running in a container")

        containers = Containers(DEFAULT_GRIDAPPSD_DOCKER_CONFIG)

        gridappsd_container = None
        try:
            gridappsd_container = client.containers.get("gridappsd")
            _log.info(f"{gridappsd_container.name} container found")
            if rebuild_if_present:
                gridappsd_container.kill()
        except docker.errors.NotFound:
            _log.debug("gridappsd container not found!")

        try:
            if gridappsd_container is None:
                containers.start()

                # the gridappsd container itself will take a bit to start up.
                time.sleep(30)

                tries = 30
                while True:
                    tries -= 1
                    if tries <=0:
                        raise RuntimeError("Couldn't connect to gridappsd server in a timely manner!")
                    try:
                        g = GridAPPSD()
                        if g.connected:
                            _log.info("Connected to gridappsd!")
                            g.disconnect()
                            break

                    except stomp.exception.ConnectFailedException or stomp.exception.NotConnectedException:
                        _log.error("Retesting connection")

            yield gridappsd_container
        finally:
            if stop_after:
                containers.stop()


    def get_docker_in_docker() -> Container:
        """
        Grab the parent container named the same as the current machine's hostname.  We are assuming that this
        is going to be a container.
        """
        # There needs to be a test to make sure that the current container (assuming run in dev environment)
        # is able to be run from the docker dev environment.  That environment is going to be assumed to be
        # the same name as the host name of the container.  See the docker-compose starting the dev environment
        hostname = str(open("/etc/hostname", "rt").read().strip())
        client = docker.from_env()
        try:
            parent_container = client.containers.get(hostname)
            _log.info(f"Inside parent container: {hostname}")
            # Setup to use gridappsd as the connection address.  This value is used in the
            # utils script to establish connection with the gridappsd server
            os.environ["GRIDAPPSD_ADDRESS"] = "gridappsd"
        except docker.errors.NotFound:
            _log.debug(f"Docker container is not named this hostname {hostname}")
            parent_container = None
        return parent_container



# -------------------------------------------------------------------------------
# Copyright (c) 2018, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------
DEFAULT_FNCS_LOCATION = 'tcp://localhost:5570'

BASE_TOPIC = '/topic/goss.gridappsd'
FNCS_BASE_INPUT_TOPIC = '/topic/goss.gridappsd.simulation.input'
FNCS_BASE_OUTPUT_TOPIC = '/topic/goss.gridappsd.simulation.output'
BASE_SIMULATION_TOPIC = '/topic/goss.gridappsd.simulation'
BASE_SIMULATION_LOG_TOPIC = "/topic/goss.gridappsd.simulation.log"

BLAZEGRAPH = "/queue/goss.gridappsd.process.request.data.powergridmodel"
# https://gridappsd.readthedocs.io/en/latest/using_gridappsd/index.html#querying-logs
LOGS = "/queue/goss.gridappsd.process.request.data.log"
# https://gridappsd.readthedocs.io/en/latest/using_gridappsd/index.html#timeseries-api
TIMESERIES = "/queue/goss.gridappsd.process.request.data.timeseries"

CONFIG = "/queue/goss.gridappsd.process.request.config"
PLATFORM_STATUS = "/queue/goss.gridappsd.process.request.status.platform"

BASE_TOPIC_PREFIX = "goss.gridappsd"
PROCESS_PREFIX = ".".join((BASE_TOPIC_PREFIX, "process"))
REQUEST_PLATFORM_STATUS = ".".join([PROCESS_PREFIX, "request.status.platform"])

REQUEST_DATA = ".".join((PROCESS_PREFIX, "request.data"))
REQUEST_SIMULATION_STATUS = ".".join((PROCESS_PREFIX, "request.status.simulation"))
REQUEST_SIMULATION = ".".join((PROCESS_PREFIX, "request.simulation"))

REQUEST_POWERGRID_DATA = ".".join((REQUEST_DATA, "powergridmodel"))
REQUEST_TIMESERIES_DATA = ".".join((REQUEST_DATA, "timeseries"))

REQUEST_REGISTER_APP = ".".join((PROCESS_PREFIX, "request.app.remote.register"))
REQUEST_APP_START = ".".join((PROCESS_PREFIX, "request.app.start"))
BASE_APPLICATION_HEARTBEAT = ".".join((BASE_TOPIC_PREFIX, "heartbeat"))


def platform_log_topic():
    """ Utility method for getting the platform.log base topic
    """
    return "/topic/{}.{}".format(BASE_TOPIC_PREFIX, "platform.log")


def service_input_topic(service_id, simulation_id):
    """ Utility method for getting the input topic for a specific service.

    The service id should be the registered service with the platform.  One
    can get the list of registered services by using the `GridAPPSD.get_platform_status()`
    method with services=True.

    Examples of service_id are as follows:
        - dnp3

    :param service_id:
    :param simulation_id:
    :return:
    """
    assert service_id, "service_id cannot be empty"
    assert simulation_id, "simulation_id cannot be empty"
    return "{}.{}.{}.input".format(BASE_SIMULATION_TOPIC, service_id, simulation_id)


def service_output_topic(service_id, simulation_id):
    """ Utility method for getting the output topic for a specific service.

    The service id should be the registered service with the platform.  One
    can get the list of registered services by using the `GridAPPSD.get_platform_status()`
    method with services=True.

    Examples of service_id are as follows:
        - dnp3

    :param service_id:
    :param simulation_id:
    :return:str: Topic to subscribe to for service specific output.
    """
    assert service_id, "Service id cannot be empty"
    assert simulation_id, "Simulation id cannot be empty"
    return "{}.{}.{}.output".format(BASE_SIMULATION_TOPIC, service_id, simulation_id)


def application_input_topic(application_id, simulation_id):
    """ Utility method for getting the input topic for a specific application.

    The application_id should be the registered service with the platform.  One
    can get the list of registered application by using the `GridAPPSD.get_platform_status()`
    method with applications=True.

    :param application_id:
    :param simulation_id:
    :return:str: Topic to publish to for a specific application.
    """
    assert application_id, "application_id cannot be empty"
    assert simulation_id, "simulation_id cannot be empty"
    return "{}.{}.{}.input".format(BASE_SIMULATION_TOPIC, application_id, simulation_id)


def application_output_topic(application_id, simulation_id):
    """ Utility method for getting the output topic for a specific application.

    The application_id should be the registered service with the platform.  One
    can get the list of registered application by using the `GridAPPSD.get_platform_status()`
    method with applications=True.

    :param application_id:
    :param simulation_id:
    :return: str: Topic to subscribe to for application specific output.
    """
    assert application_id, "application_id cannot be empty"
    #assert simulation_id, "simulation_id cannot be empty"
    if simulation_id is None:
        return "{}.{}.output".format(BASE_TOPIC, application_id)
    else:
        return "{}.{}.{}.output".format(BASE_SIMULATION_TOPIC, application_id, simulation_id)


def simulation_output_topic(simulation_id):
    """ Gets the topic for subscribing to output from the simulation.

    :param simulation_id:
    :return: str: Topic to subscribe to data from teh simulation.
    """
    return "{}.{}.{}".format(BASE_SIMULATION_TOPIC, 'output', simulation_id)


def simulation_input_topic(simulation_id):
    """ Gets the topic to write data to for the simulation

    :param simulation_id:
    :return: str: Topic to write data for the simulation.
    """
    return "{}.{}.{}".format(BASE_SIMULATION_TOPIC, 'input', simulation_id)


def simulation_log_topic(simulation_id):
    """https://gridappsd.readthedocs.io/en/latest/using_gridappsd/index.html#subscribing-to-logs
    """
    return "{}.{}".format(BASE_SIMULATION_LOG_TOPIC, simulation_id)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gridappsd']

package_data = \
{'': ['*'], 'gridappsd': ['conf/*']}

install_requires = \
['PyYAML==5.4', 'dateutils==0.6.7', 'pytz==2019.3', 'stomp.py==6.0.0']

entry_points = \
{'console_scripts': ['register_app = gridappsd.register_app:main']}

setup_kwargs = {
    'name': 'gridappsd-python',
    'version': '2.2.20220401',
    'description': 'A GridAPPS-D Python Adapter',
    'long_description': '[![Run All Pytests](https://github.com/GRIDAPPSD/gridappsd-python/actions/workflows/run-pytest.yml/badge.svg)](https://github.com/GRIDAPPSD/gridappsd-python/actions/workflows/run-pytest.yml)\n\n# gridappsd-python\nPython library for developing applications and services against the gridappsd api\n\n## Requirements\n\nThe gridappsd-python library requires a  python version >= 3.6 and < 4 in order to work properly (Note no testing\nhas been done with python 4 to date).\n\n## Installation\n\nThe recommended installation of `gridappsd-python` is in a separate virtual environment.  Executing the following\nwill create an environment called `griddapps-env`.\n\n```shell\npython3 -m venv gridappsd-env\n```\n\nSourcing the gridappsd-env activates the newly created python environment.\n\n```shell\nsource gridappsd-env/bin/activate\n```\n\nUpgrade pip to the latest (some packages require 19.0+ version of pip).\n\n```shell\npython -m pip install pip --upgrade\n```\n\nInstall the latest `gridappsd-python` and its dependencies in the virtual environment.\n\n```shell\npip install gridappsd-python\n```\n\n### Verifying things are working properly\n\nThe following code snippet assumes you have created a gridappsd instance using the steps in\nhttps://github.com/GRIDAPPSD/gridappsd-docker.\n\nCreate a test script (tester.py) with the following content.\n\n```python\n\nfrom gridappsd import GridAPPSD\n\ndef on_message_callback(header, message):\n    print(f"header: {header} message: {message}")\n\n# Note these should be changed on the server in a cyber secure environment!\nusername = "app_user"\npassword = "1234App"\n\n# Note: there are other parameters for connecting to\n# systems other than localhost\ngapps = GridAPPSD(username=username, password=password)\n\nassert gapps.connected\n\ngapps.send(\'send.topic\', {"foo": "bar"})\n\n# Note we are sending the function not executing the function in the second parameter\ngapps.subscribe(\'subscribe.topic\', on_message_callback)\n\ngapps.send(\'subcribe.topic\', \'A message about subscription\')\n\ntime.sleep(5)\n\ngapps.close()\n\n```\n\nStart up the gridappsd-docker enabled platform.  Then run the following to execute the tester.py script\n\n```shell\npython tester.py\n```\n\n## Application Developers\n\n### Deployment\n\nPlease see [DOCKER_CONTAINER.md](DOCKER_CONTAINER.md) for working within the docker application base container.\n\n### Local Development\n\nDeveloping applications against gridappsd using the `gridappsd-python` library should follow the same steps\nas above, however with a couple of environmental variables specified.  The following environmental variables are\navailable to provide the same context that would be available from inside the application docker container.  These\nare useful to know for developing your application outside of the docker context (e.g. in a python notebook).\n\n***NOTE: you can also define these your ~./bashrc file so you don\'t have to specify them all the time***\n\n```shell\n# export allows all processes started by this shell to have access to the global variable\n\n# address where the gridappsd server is running - default localhost\nexport GRIDAPPSD_ADDRESS=localhost\n\n# port to connect to on the gridappsd server (the stomp client port)\nexport GRIDAPPSD_PORT=61613\n\n# username to connect to the gridappsd server\nexport GRIDAPPSD_USER=app_user\n\n# password to connect to the gridappsd server\nexport GRIDAPPSD_PASSWORD=1234App\n\n# Note these should be changed on the server in a cyber secure environment!\n```\n\nThe following is the same tester code as above, but with the environment variables set.  The environment variables\nshould be set in your environment when running the application inside our docker container.\n\n```python\n\nfrom gridappsd import GridAPPSD\n\ndef on_message_callback(header, message):\n    print(f"header: {header} message: {message}")\n\n# Create GridAPPSD object and connect to the gridappsd server.\ngapps = GridAPPSD()\n\nassert gapps.connected\n\ngapps.send(\'send.topic\', {"foo": "bar"})\n\n# Note we are sending the function not executing the function in the second parameter\ngapps.subscribe(\'subscribe.topic\', on_message_callback)\n\ngapps.send(\'subcribe.topic\', \'A message about subscription\')\n\ntime.sleep(5)\n\ngapps.close()\n\n```\n\n## Developers\n\nThis project uses poetry to build the environment for execution.  Follow the instructions\nhttps://python-poetry.org/docs/#installation to install poetry.  As a developer I prefer not to have poetry installed\nin the same virtual environment that my projects are in.\n\nClone the github repository:\n\n```shell\ngit clone https://github.com/GRIDAPPSD/gridappsd-python -b develop\ncd gridappsd-python\n```\n\nThe following commands build and install a local wheel into an environment created just for this package.\n\n```shell\n# Build the project (stores in dist directory both .tar.gz and .whl file)\npoetry build\n\n# Install the wheel into the environment and the dev dependencies\npoetry install\n\n# Install only the library dependencies\npoetry install --no-dev\n```\n\n***Note:*** Poetry does not have a setup.py that you can install in editable mode like with pip install -e ., however\nyou can extract the generated setup.py file from the built tar.gz file in the dist directory.  Just extract the\n.tar.gz file and copy the setup.py file from the extracted directory to the root of gridappsd-python.  Then you can\nenable editing through pip install -e. as normal.\n\n\n## Testing\n\nTesting has become an integral part of the software lifecycle.  The `gridappsd-python` library has both unit and\nintegration tests available to be run.  In order to execute these, you must have installed the gridappsd-python library\nas above with dev-dependencies.\n\nDuring the testing phase the docker containers required for the tests are downloaded from\ndockerhub and started.  By default the `develop` tag is used to test the library using pytest.  \nOne can customize the docker image tag by setting the environmental\nvariable `GRIDAPPSD_TAG_ENV` either by `export GRIDAPPSD_TAG_ENV=other_tag` or by executing \npytest with the following:\n\n```shell script\n\n# Export environmental variables and all tests will use the same tag (other_tag) to pull from docker hub.\n# Default tag is develop\nexport GRIDAPPSD_TAG_ENV=other_tag\npytest\n\n# Tests also require the username and password to be avaialable as environmental variables \n# in order for them to properly run these tests\nexport GRIDAPPSD_USER=user\nexport GRIDAPPSD_PASSWORD=pass\n\npytest\n```\n\n ***NOTE: the first running the tests will download all of the docker images associated with the\n [GOSS-GridAPPS-D](http://github.com/GRIDAPPSD/GOSS-GridAPPS-D) repository.  This process may take some time.***\n \n### Running tests created in a new project\n\nThe `gridappsd-python` library exposes a testing environment through the `gridappsd.docker_handler` module.  Including the following\n`conftest.py` in the root of your base test directory allows tests to reference these.  Using these fixtures will start all of the\nbase containers required for `gridappsd` to run.  \n\n```python\n\n# conftest.py\n# Create a conftest.py file in the root of the tests directory to enable usage throughout the tests directory and below. \n# \n# Tested project structure an layout\n#\n# project-folder\\\n#   mainmodule\\\n#     __init__.py\n#     myapplication.py\n#   tests\\\n#     conftest.py\n#     test_myapplication.py\n#   README.md\n\nimport logging\nimport os\nimport sys\n\nimport pytest\n\nfrom gridappsd import GridAPPSD, GOSS\nfrom gridappsd.docker_handler import run_dependency_containers, run_gridappsd_container, Containers\n\nlevels = dict(\n    CRITICAL=50,\n    FATAL=50,\n    ERROR=40,\n    WARNING=30,\n    WARN=30,\n    INFO=20,\n    DEBUG=10,\n    NOTSET=0\n)\n\n# Get string representation of the log level passed\nLOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO")\n\n# Make sure the level passed is one of the valid levels.\nif LOG_LEVEL not in levels.keys():\n    raise AttributeError("Invalid LOG_LEVEL environmental variable set.")\n\n# Set the numeric version of log level to pass to the basicConfig function\nLOG_LEVEL = levels[LOG_LEVEL]\n\nlogging.basicConfig(stream=sys.stdout, level=LOG_LEVEL,\n                    format="%(asctime)s|%(levelname)s|%(name)s|%(message)s")\nlogging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)\nlogging.getLogger("docker.utils.config").setLevel(logging.INFO)\nlogging.getLogger("docker.auth").setLevel(logging.INFO)\n\n\nSTOP_CONTAINER_AFTER_TEST = os.environ.get(\'GRIDAPPSD_STOP_CONTAINERS_AFTER_TESTS\', True)\n\n\n@pytest.fixture(scope="module")\ndef docker_dependencies():\n    print("Docker dependencies")\n    # Containers.reset_all_containers()\n\n    with run_dependency_containers(stop_after=STOP_CONTAINER_AFTER_TEST) as dep:\n        yield dep\n    print("Cleanup docker dependencies")\n\n\n@pytest.fixture\ndef gridappsd_client(request, docker_dependencies):\n    with run_gridappsd_container(stop_after=STOP_CONTAINER_AFTER_TEST):\n        gappsd = GridAPPSD()\n        gappsd.connect()\n        assert gappsd.connected\n        models = gappsd.query_model_names()\n        assert models is not None\n        if request.cls is not None:\n            request.cls.gridappsd_client = gappsd\n        yield gappsd\n\n        gappsd.disconnect()\n\n\n@pytest.fixture\ndef goss_client(docker_dependencies):\n    with run_gridappsd_container(stop_after=STOP_CONTAINER_AFTER_TEST):\n        goss = GOSS()\n        goss.connect()\n        assert goss.connected\n\n        yield goss\n\n```\n\nUsing the above fixtures from inside a test module and test function looks like the following:\n\n```python\n\n# Example test function using the gridappsd_client fixture \n\n@mock.patch.dict(os.environ, {"GRIDAPPSD_APPLICATION_ID": "helics_goss_bridge.py"})\ndef test_gridappsd_status(gridappsd_client):\n    gappsd = gridappsd_client\n    assert "helics_goss_bridge.py" == gappsd.get_application_id()\n    assert gappsd.get_application_status() == ProcessStatusEnum.STARTING.value\n    assert gappsd.get_service_status() == ProcessStatusEnum.STARTING.value\n    gappsd.set_application_status("RUNNING")\n\n    assert gappsd.get_service_status() == ProcessStatusEnum.RUNNING.value\n    assert gappsd.get_application_status() == ProcessStatusEnum.RUNNING.value\n\n    gappsd.set_service_status("COMPLETE")\n    assert gappsd.get_service_status() == ProcessStatusEnum.COMPLETE.value\n    assert gappsd.get_application_status() == ProcessStatusEnum.COMPLETE.value\n\n    # Invalid\n    gappsd.set_service_status("Foo")\n    assert gappsd.get_service_status() == ProcessStatusEnum.COMPLETE.value\n    assert gappsd.get_application_status() == ProcessStatusEnum.COMPLETE.value\n\n```\n\n',
    'author': 'C. Allwardt',
    'author_email': 'craig.allwardt@pnnl.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gridappsd.readthedocs.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections import namedtuple

house_keys = [
    'name',
    'parent',
    'coolingSetpoint',
    'coolingSystem',
    'floorArea',
    'heatingSetpoint',
    'heatingSystem',
    'hvacPowerFactor',
    'numberOfStories',
    'thermalIntegrity',
    'id',
    'fdrid'
]
House = namedtuple('House', house_keys)

# class House(HouseBase):
#     def __dict__


class Houses:
    class __SingltonHouses:
        def __init__(self, gappsd: 'GridAPPSD'):
            self._gappsd = gappsd
            self._houses = {}

        def __str__(self):
            return repr(self) + self._gappsd

        def get_houses_for_feeder(self, feeder):
            if feeder not in self._houses:
                self._populate(feeder)
            return self._houses.get(feeder)

        def _populate(self, feeder):
            query = """# list houses - DistHouse
PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX c:  <http://iec.ch/TC57/CIM100#>
SELECT ?fdrname ?name ?parent ?coolingSetpoint ?coolingSystem ?floorArea ?heatingSetpoint ?heatingSystem ?hvacPowerFactor ?numberOfStories ?thermalIntegrity ?id ?fdrid 
WHERE { 
#	VALUES ?fdrid {"_9CE150A8-8CC5-A0F9-B67E-BBD8C79D3095"}  # R2 12.47 3
#	VALUES ?fdrid {"_4F76A5F9-271D-9EB8-5E31-AA362D86F2C3"}  # 8500-node
# VALUES ?fdrid {"_E407CBB6-8C8D-9BC9-589C-AB83FBF0826D"}  # 123 PV/Triplex
#   VALUES ?fdrid {"_503D6E20-F499-4CC7-8051-971E23D0BF79"}  # 123 Transactive
   VALUES ?fdrid {"%s"}
   ?h r:type c:House.
   ?h c:IdentifiedObject.name ?name.
   ?h c:IdentifiedObject.mRID ?id.
   ?h c:House.floorArea ?floorArea.
   ?h c:House.numberOfStories ?numberOfStories.
   OPTIONAL{?h c:House.coolingSetpoint ?coolingSetpoint.}
   OPTIONAL{?h c:House.heatingSetpoint ?heatingSetpoint.}
   OPTIONAL{?h c:House.hvacPowerFactor ?hvacPowerFactor.}
   ?h c:House.coolingSystem ?coolingSystemRaw.
   	bind(strafter(str(?coolingSystemRaw),"HouseCooling.") as ?coolingSystem) 
   ?h c:House.heatingSystem ?heatingSystemRaw.
   	bind(strafter(str(?heatingSystemRaw),"HouseHeating.") as ?heatingSystem)
   ?h c:House.thermalIntegrity ?thermalIntegrityRaw.
   	bind(strafter(str(?thermalIntegrityRaw),"HouseThermalIntegrity.") as ?thermalIntegrity)
   ?h c:House.EnergyConsumer ?econ.
   ?econ c:IdentifiedObject.name ?parent.
   ?fdr c:IdentifiedObject.mRID ?fdrid.
   ?fdr c:IdentifiedObject.name ?fdrname.
   ?econ c:Equipment.EquipmentContainer ?fdr.
} ORDER BY ?fdrname ?name
""" % feeder
            response = self._gappsd.query_data(query)
            houses = {}
            for rec in response['data']['results']['bindings']:
                create_order = {}
                name = None
                for d in house_keys:
                    if d == 'name':
                        name = rec[d]['value']
                    try:
                        create_order[d] = rec[d]['value']
                    except KeyError:
                        create_order[d] = None

                houses[name] = House(**create_order)

            self._houses[feeder] = houses


    instance = None

    def __init__(self, gappsd):
        if not Houses.instance:
            Houses.instance = Houses.__SingltonHouses(gappsd)
        else:
            Houses.instance._gappsd = gappsd

    def __getattr__(self, name):
        return getattr(self.instance, name)




import time
from copy import deepcopy
import json
import logging

from . import topics as t
from .topics import simulation_input_topic

_log = logging.getLogger(__name__)


class SimulationFailedToStartError(Exception):
    """Exception raised if a simulation fails to start."""
    pass


class Simulation(object):
    """ Simulation object allows controlling simulations through a python API.

    The simulation object allows controlling and monitoring of simulations through
    a python API.  It is capable of starting, stopping, pausing and restarting simulations
    that are run on GridAPPSD.

    There are four events that can be registered: ontimestep, onmeasurement, oncomplete, and
    onstart.  To register one of the events call the method add_ontimestep_callback,
    add_onmeasurement_callback, add_oncomplete_callback or add_onstart_callback method respectively.
    """
    def __init__(self, gapps: 'GridAPPSD', run_config):
        assert type(gapps).__name__ == 'GridAPPSD',  "Must be an instance of GridAPPSD"
        assert isinstance(run_config, dict)
        self._gapps = gapps
        self._run_config = deepcopy(run_config)
        self._running_or_paused = False

        # Will be populated when the simulation is first started.
        self.simulation_id = None

        self.__on_start = set()
        self.__on_next_timestep_callbacks = set()
        self.__on_simulation_complete_callbacks = set()

        self._measurement_count = 0
        self._log_count = 0
        self._platform_log_count = 0

        self._num_timesteps = round(float(run_config['simulation_config']['duration']))

        # Devices that the user wants measurements from
        self._device_measurement_filter = {}

        self.__filterable_measurement_callback_set = set()

    def start_simulation(self, timeout=30):
        """ Start the configured simulation by calling the REQUEST_SIMULATION endpoint.
        """
        resp = self._gapps.get_response(t.REQUEST_SIMULATION, json.dumps(self._run_config), timeout=timeout)

        if 'simulationId' not in resp:
            message = "Simulation was not able to run\n" + str(resp)
            raise SimulationFailedToStartError(message)

        self._running_or_paused = True
        self.simulation_id = resp['simulationId']

        # Subscribe to the different components necessary to run and receive
        # simulated measurements and messages.
        self._gapps.subscribe(t.simulation_log_topic(self.simulation_id), self.__on_simulation_log)
        self._gapps.subscribe(t.simulation_output_topic(self.simulation_id), self.__onmeasurement)
        self._gapps.subscribe(t.platform_log_topic(), self.__on_platformlog)

        for p in self.__on_start:
            p(self)

    def pause(self):
        """ Pause simulation"""
        _log.debug("Pausing simulation")
        command = dict(command="pause")
        self._gapps.send(simulation_input_topic(self.simulation_id), json.dumps(command))
        self._running_or_paused = True

    def stop(self):
        """ Stop the simulation"""
        _log.debug("Stopping simulation")
        command = dict(command="stop")
        self._gapps.send(simulation_input_topic(self.simulation_id), json.dumps(command))
        self._running_or_paused = True

    def resume(self):
        """ Resume the simulation"""
        _log.debug("Resuming simulation")
        command = dict(command="resume")
        self._gapps.send(simulation_input_topic(self.simulation_id), json.dumps(command))
        self._running_or_paused = True

    def run_loop(self):
        """ Loop around the running of the simulation itself.

        Example:

            gapps = GridAPPSD()

            # Create simulation object.
            simulation = Simulation(gapps, config)
            simulation.add_ontimestep_callback(ontimestep)
            simulation.add_oncomplete_callback(oncomplete)
            simulation.add_onmeasurement_callback(onmeasurment)
            simulation.add_onstart_callback(onstart)
            simulation.run_loop()

        """
        if not self._running_or_paused:
            _log.debug("Running simulation in loop until simulation is done.")
            self.start_simulation()

        while self._running_or_paused:
            time.sleep(0.01)

    def resume_pause_at(self, pause_in):
        """ Resume the simulation and have it automatically pause after specified amount of seconds later.
        
        :param pause_in: number of seconds to run before pausing the simulation
        """
        _log.debug("Resuming simulation. Will pause after {} seconds".format(pause_in))
        command = dict(command="resumePauseAt", input=dict(pauseIn=pause_in))
        self._gapps.send(simulation_input_topic(self.simulation_id), json.dumps(command))
        self._running_or_paused = True

    def add_onmeasurement_callback(self, callback, device_filter=()):
        """ registers an onmeasurment callback to be called when measurements have come through.

        Note:

            The device_filter has not been fully implemented at present!

        The callback parameter must be a function that takes three arguments (the simulation object,
        a timstep and a measurement dictionary)

        Callback Example:

            def onmeasurment(sim, timestep, measurements):
                print("timestep: {}, measurement: {}".format(timestep, measurements))

        :param callback: Function to be called during running of simulation
        :param device_filter: Future filter of measurements
        :return:
        """
        self.__filterable_measurement_callback_set.add((callback, device_filter))

    def add_onstart_callback(self, callback):
        """ registers a start callback that is called when the simulation is started

        Callback Example:

            def onstart(sim):
                print("Sim started: {}".format(sim.simulation_id))



        :param callback:
        :return:
        """
        self.__on_start.add(callback)

    def add_oncomplete_callback(self, callback):
        """ registers a completion callback when the last timestep has been requested.

        Callback Example:

            def onfinishsimulation(sim):
                print("Completed simulator")

        :param callback:
        :return:
        """
        self.__on_simulation_complete_callbacks.add(callback)

    def add_ontimestep_callback(self, callback):
        """ register a timestep callback

        Callback Example:

            def ontimestep(sim, timestep):
                print("Timestamp: {}".format(timestep))

        :param callback:
        :return:
        """
        self.__on_next_timestep_callbacks.add(callback)

    def __on_platformlog(self, headers, message):
        try:
            if self.simulation_id == message['processId']:
                _log.debug(f"__on_platform_log: message: {message}")
        except KeyError as e:
            _log.error(f"__on_platformlog keyerror({e}): {message}")

        if 'command' in message:
            _log.debug("Command was: {}".format(message))

    def __on_simulation_log(self, headers, message):
        # Handle the callbacks here
        if 'logMessage' in message:
            log_message = message['logMessage']
            # if this is the last timestamp then call the finished callbacks
            if log_message == f"Simulation {self.simulation_id} complete":
                for p in self.__on_simulation_complete_callbacks:
                    p(self)
                self._running_or_paused = False
                _log.debug("Simulation completed")
            elif log_message.startswith("incrementing to "):
                timestep = log_message[len("incrementing to "):]
                for p in self.__on_next_timestep_callbacks:
                    p(self, int(timestep))

    def __onmeasurement(self, headers, message):
        """ Call the measurement callbacks

        :param headers:
        :param message:
        :return:
        """
        sim_id = message['simulation_id']
        timestamp = message['message']['timestamp']
        measurements = message['message']['measurements']
        for p in self.__filterable_measurement_callback_set:
            p[0](self, timestamp, measurements)



# -------------------------------------------------------------------------------
# Copyright (c) 2018, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import calendar
from uuid import uuid4
import time

import pytz


class DifferenceBuilder(object):
    """ Automates the building of forward and reverse cim differences

    """

    def __init__(self, simulation_id):

        self._simulation_id = simulation_id
        self._forward = []
        self._reverse = []

    def add_difference(self, object_id, attribute, forward_value, reverse_value):
        """ Add forward and reverse unit for an object attribute.

         All of the parameters must be serializable for sending the GOSS message bus.
         """
        forward = dict(object=object_id, attribute=attribute, value=forward_value)
        reverse = dict(object=object_id, attribute=attribute, value=reverse_value)
        self._forward.append(forward)
        self._reverse.append(reverse)

    def clear(self):
        self._forward = []
        self._reverse = []

    def get_message(self, epoch = None):
        if epoch is None:
            epoch = calendar.timegm(time.gmtime())
        msg = dict(command="update",
                   input=dict(simulation_id=self._simulation_id,
                              message=dict(timestamp=epoch,
                                           difference_mrid=str(uuid4()),
                                           reverse_differences=self._reverse,
                                           forward_differences=self._forward)))
        return msg.copy()
